#!/usr/bin/env python
#
#
#   Copyright (c) 2007 Jesse Guardiani <jesse@guardiani.us>
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License as
#   published by the Free Software Foundation; either version 2 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
#   02111-1307, USA.
#

import pygame, time, gc, gobject, sys, os
from pygame.locals import *
import globals,widgets
from db import db as db
from manager import manager as manager
from theme   import theme   as theme


class BaseView():
  def __init__(self, caption=None, show_clear_playlist=True):
    ''' This class defines a basic interface for a view (or in web programming terms, a page).
    If Manager is the brain of the app, then each View is the brain of that screen or page. '''
    self.widgets = [] # all widgets for a view are stored within that view
    self.scroll_widget = None
    self.is_temporary = False # can this view be garbage collected?
    self.back_button  = None
    self.add_to_history = True
    self._load_back_button()
    self.base_combo_timeout_start = 0
    self.empty_view = False
    self.title_label = None
    self.caption     = caption
    if caption!=None:
      self.title_label = widgets.Label(caption, Rect(0,0,125,26), 4, 4, 22)
      self.widgets.append(self.title_label)
    if show_clear_playlist: self.widgets.append(widgets.ClearPlaylistButton())

  def _load_back_button(self):
    if self.back_button or self.back_button in self.widgets: return False
    self.back_button = widgets.BackButton()
    self.widgets.append(self.back_button)

  def get_widgets(self):
    return self.widgets
  
  def handle_event(self,event):
    if self.scroll_widget!=None and manager.nowplaying_button!=None and manager.nowplaying_button.is_pressed():
      if event.type == KEYDOWN and (event.key == K_LEFT or event.key == K_RIGHT):
        can_handle = (time.time()-self.base_combo_timeout_start > 0.1)
        if can_handle==True and event.key == K_LEFT:
          self.scroll_widget.move_home()
        elif can_handle==True and event.key == K_RIGHT:
          self.scroll_widget.move_end()

        manager.nowplaying_button.set_ignore_next_mouseup(True)
        self.base_combo_timeout_start = time.time()
        return True

    for widget in self.widgets:
      if widget.handle_event(event): return True
    return False

  def show(self):
    if manager.view != self:
      manager.loading_sign()

      if manager.view:
        if self not in manager.view_history:
          if manager.view.add_to_history:
            manager.view_history.append(manager.view)
        else:
          # roll back history to this view before this view
          index = manager.view_history.index(self)
          print "rolling back view history to %s,index=%s" % (self,index)
          manager.back(index)
          #manager.view_history = manager.view_history[:index]
      manager.view  = self
      manager.dirty = True
      self.back_button._load_image()
#      print "view_history=%s,view=%s" % (manager.view_history,manager.view)
      pygame.event.clear()
      if self.scroll_widget:
        self.scroll_widget.snap_selected()
        self.scroll_widget.scrollbar_toggle(manager.scrollbar_state)

  def update(self):
    for widget in self.widgets: widget.update()

  def close(self):
    self.scroll_widget = None
    for widget in self.widgets: widget.close()
    self.widgets = None


class TimerView(BaseView):
  def __init__(self,timeout=3000,caption=None):
    ''' This is the Zoomed Album Cover view '''
    BaseView.__init__(self,caption)
    self.timeout        = timeout # jump to last_view after timeout milliseconds
    self.timeout_cb     = None
    self.add_to_history = False # TimerViews don't show up in the view history

  def show(self):
    BaseView.show(self)
    self.reset_timeout()

  def reset_timeout(self):
    if self.timeout_cb: gobject.source_remove(self.timeout_cb)
    self.timeout_cb = gobject.timeout_add(self.timeout, self.show_last_view)

  def stop_timeout(self):
    if self.timeout_cb: gobject.source_remove(self.timeout_cb)
    self.timeout_cb = None

  def show_last_view(self):
    self.stop_timeout()

    if manager.view != self:
      return False # timeout invalid when we aren't the current view

    manager.back()


class AlbumView(BaseView):
  def __init__(self):
    ''' This is the album selection view '''
    BaseView.__init__(self, 'Albums')
    self.combo_timeout_start = 0
    self.album_widget = None
    self._load_album_widget()
    self.scroll_widget = self.album_widget
    self.widgets.append(widgets.VolumeButton())

  def _load_album_widget(self):
    if self.album_widget: return

    manager.clear()
    status_area = widgets.Text("LOADING ALBUM ART")
    status_area.rect.topleft = (130,200)
    status_area.update()
    status_area.draw(manager.screen)
    pygame.display.flip()

    widget = widgets.ScrollWidget(Rect(
      130,
      manager.screen.get_rect().top,
      545,
      manager.screen.get_rect().bottom
    ),self._make_ok_button_callback(),
    self._get_album_list(),
    self.albumartist_generator_callback)
    
    self.album_widget = widget
    self.widgets.append(widget)

  def _get_album_list(self):
    processtart=time.time()
    list =  db.get_albums()
    print "album query took:" + str(time.time()-processtart)
    return list

  def albumartist_generator_callback(self,id,row):
    return widgets.AlbumArtist(row['album_id'],row['album'],row['artist_id'],row['artist'])

  def _make_ok_button_callback(self):
    def _ok_button_callback():
      sel_sprite = manager.view.scroll_widget.get_focused_sprite()
      return manager.show_songs(sel_sprite)
    return _ok_button_callback

  def handle_event(self,event):
    if manager.nowplaying_button!=None and manager.nowplaying_button.is_pressed():
      if event.type == KEYDOWN and (event.key==K_F4):
        can_handle = (time.time()-self.combo_timeout_start > 0.1)

        if can_handle==True and event.key == K_F4 and self.scroll_widget.has_focused():
          i = self.scroll_widget.get_focused_sprite_index()
          sel_albumartist = self.scroll_widget.visible_sprites[i]['sprite']
          print "artist = ",sel_albumartist.artist.name
          manager.ignore_keypad_till = time.time() + 2
          manager.nowplaying_button.set_pressed(False) # will change views below
          manager.show_artist_songs(sel_albumartist.artist_id,sel_albumartist.artist.name)

        manager.nowplaying_button.set_ignore_next_mouseup(True)
        self.combo_timeout_start = time.time()
        return True

    return BaseView.handle_event(self,event)


class ArtistView(BaseView):
  def __init__(self):
    ''' This is the artist selection view '''
    BaseView.__init__(self, 'Artists')
    self.combo_timeout_start = 0
    self.album_widget = None
    self._load_album_widget()
    self.scroll_widget = self.album_widget
    self.widgets.append(widgets.VolumeButton())

  def _load_album_widget(self):
    if self.album_widget: return

    manager.clear()
    status_area = widgets.Text("LOADING ARTISTS")
    status_area.rect.topleft = (130,200)
    status_area.update()
    status_area.draw(manager.screen)
    pygame.display.flip()

    widget = widgets.ScrollWidget(Rect(
      130,
      manager.screen.get_rect().top,
      545,
      manager.screen.get_rect().bottom
    ),self._make_ok_button_callback(),
    self._get_artist_list(),
    self.artist_generator_callback)
    
    self.album_widget = widget
    self.widgets.append(widget)

  def _get_artist_list(self):
    processtart=time.time()
    list =  db.get_artists()
    print "artist query took:" + str(time.time()-processtart)
    return list

  def artist_generator_callback(self,id,row):
    return widgets.Artist(row['artist_id'],row['artist'])

  def _make_ok_button_callback(self):
    def _ok_button_callback():
      sel_sprite = manager.view.scroll_widget.get_focused_sprite()
      return manager.show_artist_albums(sel_sprite)
    return _ok_button_callback

  def handle_event(self,event):
    if manager.nowplaying_button!=None and manager.nowplaying_button.is_pressed():
      if event.type == KEYDOWN and (event.key==K_F4):
        can_handle = (time.time()-self.combo_timeout_start > 0.1)

        if can_handle==True and event.key == K_F4 and self.scroll_widget.has_focused():
          i = self.scroll_widget.get_focused_sprite_index()
          sel_artist = self.scroll_widget.visible_sprites[i]['sprite']
          print "artist = ",sel_artist.artist.name
          manager.ignore_keypad_till = time.time() + 2
          manager.nowplaying_button.set_pressed(False) # will change views below
          manager.show_artist_songs(sel_artist.artist_art.id,sel_artist.artist.name)

        manager.nowplaying_button.set_ignore_next_mouseup(True)
        self.combo_timeout_start = time.time()
        return True

    return BaseView.handle_event(self,event)


class PlaylistView(BaseView):
  def __init__(self):
    ''' This is the playlist selection view '''
    BaseView.__init__(self, 'Playlists')
    self.album_widget = None
    self._load_album_widget()
    self.scroll_widget = self.album_widget
    self.widgets.append(widgets.VolumeButton())

  def _load_album_widget(self):
    if self.album_widget: return

    manager.clear()
    status_area = widgets.Text("LOADING PLAYLISTS")
    status_area.rect.topleft = (130,200)
    status_area.update()
    status_area.draw(manager.screen)
    pygame.display.flip()

    widget = widgets.ScrollWidget(Rect(
      130,
      manager.screen.get_rect().top,
      545,
      manager.screen.get_rect().bottom
    ),self._callback,
    self._get_playlist_list(),
    self.playlist_generator_callback)
    
    self.album_widget = widget
    self.widgets.append(widget)

  def _get_playlist_list(self):
    processtart=time.time()
    list = db.get_m3us()
    print "m3u query took:" + str(time.time()-processtart)
    return list

  def playlist_generator_callback(self,id,row):
    print "row=%s" % (row,)
    return widgets.M3U(row['id'],row['name'],row['path'])

  def _callback(self):
    sel_sprite = manager.view.scroll_widget.get_focused_sprite()
    return manager.show_m3u_songs(sel_sprite)


class SongView(BaseView):
  def __init__(self,album_id,album):
    ''' This is the song selection view '''
    BaseView.__init__(self, 'Songs',show_clear_playlist=False)
    self.is_temporary = True # can this view be garbage collected?
    self.combo_timeout_start = 0
    self.album       = album
    self.album_id    = album_id
    self._load_songalbum_button()
    self._load_song_widget()
    self.widgets.append(widgets.AddAlbumButton())

  def _load_song_widget(self):
    if self.scroll_widget != None:
      self.widgets.append(self.scroll_widget)
      return

    manager.clear()
    status_area = widgets.Text("LOADING SONGS")
    status_area.rect.topleft = (130,200)
    status_area.update()
    status_area.draw(manager.screen)
    pygame.display.flip()

    widget = widgets.ScrollWidget(Rect(
      130,
      manager.screen.get_rect().top,
      545,
      manager.screen.get_rect().bottom
    ),self._make_ok_button_callback(),
    self._get_song_list(),
    self.song_generator_callback)

    self.scroll_widget = widget
    self.widgets.append(widget)

  def _get_song_list(self):
    processtart=time.time()
    list = db.get_songs(album_id=self.album_id)
    print "song query took:" + str(time.time()-processtart)
    return list

  def song_generator_callback(self,id,song):
    return widgets.Song(song['track'],song['title'],song['path'],flags=song['flags'],artist=song['artist'],artist_id=song['artist_id'])

  def _play_continuous(self):
    manager.playlist.clear()
    for row in manager.view.scroll_widget.get_list():
      manager.playlist.add(row['path'])
    if manager.view.scroll_widget.has_focused(): # paranoia
      manager.playlist.play(manager.view.scroll_widget.get_focused_sprite_index())
      manager.show_nowplaying()

  def _make_ok_button_callback(self):
    def _ok_button_callback():
      sel_spr = manager.view.scroll_widget.get_focused_sprite()
      if not sel_spr: return
      if not manager.enqueue_mode:
        self._play_continuous()
      else:
        print "adding "+sel_spr.path
        manager.playlist.add(sel_spr.path)
        manager.playlist.update()
    return _ok_button_callback

  def _load_songalbum_button(self):
    ''' This is the album picture in the upper left.
    Clicking it should (eventually) bring up a hi res version of
    the album art. '''
    image = theme.get_surface((125,160))
    print "songalbum: " + self.album
    sprite = widgets.Album(self.album_id, self.album)
    sprite.draw(image,(0,0))

    artist = db.artist_of_album(self.album_id)
    globals.gprint(artist,image,(2,132),16)
    button = widgets.Button(image,self._make_songalbum_button_callback())
    button.rect.left = 0
    button.rect.top  = 26
    self.widgets.append(button)

  def _make_songalbum_button_callback(self):
    def _songalbum_button_callback():
      print "show AlbumCoverView"
      return manager.show_albumcover_view(album_id=self.album_id)
    return _songalbum_button_callback

  def handle_event(self,event):
    if manager.nowplaying_button!=None and manager.nowplaying_button.is_pressed():
      if event.type == KEYDOWN and (event.key==K_F4):
        can_handle = (time.time()-self.combo_timeout_start > 0.1)

        if can_handle==True and event.key == K_F4 and self.scroll_widget.has_focused():
          i = self.scroll_widget.get_focused_sprite_index()
          sel_song = self.scroll_widget.visible_sprites[i]['sprite']
          print "artist = ",sel_song.artist
          manager.ignore_keypad_till = time.time() + 2
          manager.nowplaying_button.set_pressed(False) # will change views below
          manager.show_artist_albums_ext(sel_song.artist_id,sel_song.artist)

        manager.nowplaying_button.set_ignore_next_mouseup(True)
        self.combo_timeout_start = time.time()
        return True

    return BaseView.handle_event(self,event)


class M3UView(BaseView):
  def __init__(self,m3u_id,m3u,m3u_path):
    ''' This is the song selection view '''
    BaseView.__init__(self, 'Playlist Songs',show_clear_playlist=False)
    self.is_temporary = True # can this view be garbage collected?
    self.combo_timeout_start = 0
    self.m3u       = m3u
    self.m3u_id    = m3u_id
    self.m3u_path  = m3u_path
    self._load_song_widget()
    if len(self.songlist)>0:
      self.widgets.append(widgets.AddAlbumButton())

  def _load_song_widget(self):
    if self.scroll_widget != None:
      self.widgets.append(self.scroll_widget)
      return

    manager.clear()
    status_area = widgets.Text("LOADING PLAYLIST SONGS")
    status_area.rect.topleft = (130,200)
    status_area.update()
    status_area.draw(manager.screen)
    pygame.display.flip()

    self.songlist  = self.load_songs()

    widget = widgets.ScrollWidget(Rect(
      130,
      manager.screen.get_rect().top,
      545,
      manager.screen.get_rect().bottom
    ),self._callback,
    self._get_song_list(),
    self.song_generator_callback)

    self.scroll_widget = widget
    self.widgets.append(widget)

  def _get_song_list(self):
    return self.songlist

  def load_songs(self):
    from m3u     import m3u     as m3u
    processtart=time.time()
    list = m3u.load(filename=self.m3u_path)
    print "m3u load took:" + str(time.time()-processtart)
    processtart=time.time()
    path = os.path.dirname(self.m3u_path)
    safelist = []
    for s in list:
      rows = db.song_of_path(s)
      if (len(rows)>0):
        safelist.append(rows[0])
    print "safelist generation took:" + str(time.time()-processtart)
    return safelist

  def song_generator_callback(self,id,song):
    return widgets.Song(song['track'],song['title'],song['path'],flags=song['flags'],artist=song['artist'],artist_id=song['artist_id'])

  def _play_continuous(self):
    manager.playlist.clear()
    for path in manager.view.scroll_widget.get_list():
      manager.playlist.add(path)
    if manager.view.scroll_widget.has_focused(): # paranoia
      manager.playlist.play(manager.view.scroll_widget.get_focused_sprite_index())
      manager.show_nowplaying()

  def _callback(self):
    sel_spr = manager.view.scroll_widget.get_focused_sprite()
    if not sel_spr: return
    if not manager.enqueue_mode:
      self._play_continuous()
    else:
      print "adding "+sel_spr.path
      manager.playlist.add(sel_spr.path)
      manager.playlist.update()

  def handle_event(self,event):
    if manager.nowplaying_button!=None and manager.nowplaying_button.is_pressed():
      if event.type == KEYDOWN and (event.key==K_F4):
        can_handle = (time.time()-self.combo_timeout_start > 0.1)

        if can_handle==True and event.key == K_F4 and self.scroll_widget.has_focused():
          i = self.scroll_widget.get_focused_sprite_index()
          sel_song = self.scroll_widget.visible_sprites[i]['sprite']
          print "artist = ",sel_song.artist
          manager.ignore_keypad_till = time.time() + 2
          manager.nowplaying_button.set_pressed(False) # will change views below
          manager.show_artist_albums_ext(sel_song.artist_id,sel_song.artist)

        manager.nowplaying_button.set_ignore_next_mouseup(True)
        self.combo_timeout_start = time.time()
        return True

    return BaseView.handle_event(self,event)


class AllSongsView(BaseView):
  def __init__(self):
    ''' This is the all songs view '''
    BaseView.__init__(self, 'All Songs',show_clear_playlist=False)
    self.album_widget = None
    self._load_album_widget()
    self.scroll_widget = self.album_widget
    self.widgets.append(widgets.VolumeButton())
    self.widgets.append(widgets.AddAlbumButton())
    self.is_temporary = True
    self.combo_timeout_start = 0

  def _load_album_widget(self):
    if self.album_widget: return

    manager.clear()
    status_area = widgets.Text("LOADING ALL SONGS")
    status_area.rect.topleft = (130,200)
    status_area.update()
    status_area.draw(manager.screen)
    pygame.display.flip()

    widget = widgets.ScrollWidget(Rect(
      130,
      manager.screen.get_rect().top,
      545,
      manager.screen.get_rect().bottom
    ),self._make_ok_button_callback(),
    self._get_all_songs_list(),
    self.song_generator_callback)
    
    self.album_widget = widget
    self.widgets.append(widget)

  def _get_all_songs_list(self):
    processtart=time.time()
    list = db.get_songs()
    print "all songs query took:" + str(time.time()-processtart)
    return list

  def song_generator_callback(self,id,song):
    track=''
    if song['track']: track = "[%s] " % (song['track'],)
    return widgets.Song(id+1,track + song['title'],song['path'],artist=song['artist'],flags=song['flags'],artist_id=song['artist_id'])

  def _make_ok_button_callback(self):
    def _ok_button_callback():
      sel_spr = manager.view.scroll_widget.get_focused_sprite()
      if not sel_spr: return
      if not manager.enqueue_mode:
        self._play_continuous()
      else:
        print "adding "+sel_spr.path
        manager.playlist.add(sel_spr.path)
        manager.playlist.update()
    return _ok_button_callback

  def handle_event(self,event):
    if manager.nowplaying_button!=None and manager.nowplaying_button.is_pressed():
      if event.type == KEYDOWN and (event.key==K_F4):
        can_handle = (time.time()-self.combo_timeout_start > 0.1)

        if can_handle==True and event.key == K_F4 and self.scroll_widget.has_focused():
          i = self.scroll_widget.get_focused_sprite_index()
          song = self.scroll_widget.visible_sprites[i]['sprite']
          print "artist = ",song.artist
          manager.ignore_keypad_till = time.time() + 2
          manager.nowplaying_button.set_pressed(False) # will change views below
          manager.show_artist_albums_ext(song.artist_id,song.artist)

        manager.nowplaying_button.set_ignore_next_mouseup(True)
        self.combo_timeout_start = time.time()
        return True

    return BaseView.handle_event(self,event)


class ArtistSongsView(BaseView):
  def __init__(self,artist_id,artist):
    ''' This is the artist songs view '''
    BaseView.__init__(self, 'Songs of Artist',show_clear_playlist=False)
    self.artist = artist
    self.artist_id = artist_id
    self.album_widget = None
    self._load_album_widget()
    self._load_songartist_button()
    self.scroll_widget = self.album_widget
    self.widgets.append(widgets.AddAlbumButton())
    self.is_temporary = True
    self.combo_timeout_start = 0

  def _load_album_widget(self):
    if self.album_widget: return

    manager.clear()
    status_area = widgets.Text("LOADING ARTIST SONGS")
    status_area.rect.topleft = (130,200)
    status_area.update()
    status_area.draw(manager.screen)
    pygame.display.flip()

    widget = widgets.ScrollWidget(Rect(
      130,
      manager.screen.get_rect().top,
      545,
      manager.screen.get_rect().bottom
    ),self._make_ok_button_callback(),
    self._get_artist_songs_list(),
    self.song_generator_callback)
    
    self.album_widget = widget
    self.widgets.append(widget)

  def _get_artist_songs_list(self):
    processtart=time.time()
    list = db.get_songs(artist_id=self.artist_id)
    print "artist songs query took:" + str(time.time()-processtart)
    return list

  def song_generator_callback(self,id,song):
    track=''
    if song['track']: track = "[%s] " % (song['track'],)
    return widgets.Song(id+1,track + song['title'],song['path'],artist=song['artist'],flags=song['flags'],artist_id=song['artist_id'],album=song['album'],album_id=song['album_id'],subtext_album=True)

  def _make_ok_button_callback(self):
    def _ok_button_callback():
      sel_spr = manager.view.scroll_widget.get_focused_sprite()
      if not sel_spr: return
      if not manager.enqueue_mode:
        self._play_continuous()
      else:
        print "adding "+sel_spr.path
        manager.playlist.add(sel_spr.path)
        manager.playlist.update()
    return _ok_button_callback

  def _load_songartist_button(self):
    ''' This is the artist picture in the upper left.
    Clicking it should bring up a hi res version of
    the album art. '''
    image = theme.get_surface((125,160))
    print "songartist: " + self.artist
    sprite = widgets.ArtistArt(self.artist_id, self.artist)
    sprite.draw(image,(0,0))

    globals.gprint(self.artist,image,(2,132),16)
    button = widgets.Button(image,self._songartist_button_callback)
    button.rect.left = 0
    button.rect.top  = 26
    self.widgets.append(button)

  def _songartist_button_callback(self):
      print "show ArtistCoverView"
      return manager.show_artistcover_view(self.artist_id)

  def handle_event(self,event):
    if manager.nowplaying_button!=None and manager.nowplaying_button.is_pressed():
      if event.type == KEYDOWN and (event.key==K_F4):
        can_handle = (time.time()-self.combo_timeout_start > 0.1)

        if can_handle==True and event.key == K_F4 and self.scroll_widget.has_focused():
          i = self.scroll_widget.get_focused_sprite_index()
          song = self.scroll_widget.visible_sprites[i]['sprite']
          print "album = ",song.album
          manager.ignore_keypad_till = time.time() + 2
          manager.nowplaying_button.set_pressed(False) # will change views below
          manager.show_songs_ext(song.album_id,song.album)

        manager.nowplaying_button.set_ignore_next_mouseup(True)
        self.combo_timeout_start = time.time()
        return True

    return BaseView.handle_event(self,event)


class ArtistAlbumView(BaseView):
  def __init__(self,artist_id,artist_name):
    ''' This is the artist songs view '''
    BaseView.__init__(self, 'Albums of Artist',show_clear_playlist=False)
    self.artist       = artist_name
    self.artist_id    = artist_id
    self.list         = []
    processtart=time.time()
    self.list =  db.get_albums(artist_id=self.artist_id)
    print "album query took:" + str(time.time()-processtart)
    self.empty_view = (len(self.list)==0)
    if self.empty_view:
      return
    self.album_widget = None
    self._load_songartist_button()
    self._load_album_widget()
    self.scroll_widget = self.album_widget
    self.widgets.append(widgets.AddAlbumButton())
    self.is_temporary = True
    self.combo_timeout_start = 0

  def _load_album_widget(self):
    if self.album_widget: return

    manager.clear()
    status_area = widgets.Text("LOADING ARTIST SONGS")
    status_area.rect.topleft = (130,200)
    status_area.update()
    status_area.draw(manager.screen)
    pygame.display.flip()

    widget = widgets.ScrollWidget(Rect(
      130,
      manager.screen.get_rect().top,
      545,
      manager.screen.get_rect().bottom
    ),self._ok_button_callback,
    self._get_artist_album_list(),
    self.album_generator_callback)
    
    self.album_widget = widget
    self.widgets.append(widget)

  def _get_artist_album_list(self):
#    processtart=time.time()
#    list =  db.get_albums(artist_id=self.artist_id)
#    print "album query took:" + str(time.time()-processtart)
    return self.list

  def album_generator_callback(self,id,row):
    return widgets.AlbumArtist(row['album_id'],row['album'],row['artist_id'],row['artist'])

  def _ok_button_callback(self):
    sel_sprite = manager.view.scroll_widget.get_focused_sprite()
    return manager.show_songs(sel_sprite)

  def _load_songartist_button(self):
    ''' This is the artist picture in the upper left.
    Clicking it should bring up a hi res version of
    the album art. '''
    image = theme.get_surface((125,160))
    print "songartist: " + self.artist
    sprite = widgets.ArtistArt(self.artist_id, self.artist)
    sprite.draw(image,(0,0))

    globals.gprint(self.artist,image,(2,132),16)
    button = widgets.Button(image,self._songartist_button_callback)
    button.rect.left = 0
    button.rect.top  = 26
    self.widgets.append(button)

  def _songartist_button_callback(self):
      print "show ArtistCoverView"
      return manager.show_artistcover_view(self.artist_id)

  def handle_event(self,event):
    if manager.nowplaying_button!=None and manager.nowplaying_button.is_pressed():
      if event.type == KEYDOWN and (event.key==K_F4):
        can_handle = (time.time()-self.combo_timeout_start > 0.1)

        if can_handle==True and event.key == K_F4:
          manager.ignore_keypad_till = time.time() + 2
          manager.nowplaying_button.set_pressed(False) # will change views below
          manager.show_artist_songs(self.artist_id,self.artist)

        manager.nowplaying_button.set_ignore_next_mouseup(True)
        self.combo_timeout_start = time.time()
        return True

    return BaseView.handle_event(self,event)


class NowPlayingView(BaseView):
  def __init__(self):
    ''' This is the nowplaying selection view '''
    BaseView.__init__(self, 'Now Playing')
    self.album = None
    self.dirty = False
    self.combo_timeout_start = 0
    self.last_selected_index = None
    self._load_scroll_widget()
    self.widgets.append(widgets.VolumeButton())
    self.widgets.append(widgets.DeleteOneButton())
    self.update()

  def _load_scroll_widget(self):
    if self.scroll_widget != None:
      self.widgets.remove(self.scroll_widget)
      self.scroll_widget = None
    gc.collect()

    manager.clear()
    if self.dirty:
      status_area = widgets.Text("ADDING TO PLAYLIST")
    else:
      status_area = widgets.Text("LOADING NOW PLAYING")
    status_area.rect.topleft = (130,200)
    status_area.update()
    status_area.draw(manager.screen)
    pygame.display.flip()

    widget = widgets.ScrollWidget(Rect(
      130,
      manager.screen.get_rect().top,
      545,
      manager.screen.get_rect().bottom
    ),self._callback,
    self._get_song_list(),
    self.song_generator_callback,
    wrap=False)

    self.scroll_widget = widget
    self.widgets.append(widget)
    self.dirty = False

  def _get_song_list(self):
    return manager.playlist.list

  def song_generator_callback(self,id,path):
    song = db.song_of_path(path)[0]
    track=''
    if song['track']: track = "[%s] " % (song['track'],)
    return widgets.Song(id+1,track + song['title'],song['path'],artist=song['artist'],flags=song['flags'],artist_id=song['artist_id'])

  def _play_continuous(self):
    if not manager.view.scroll_widget.get_focused_sprite(): return
    if manager.view.scroll_widget.has_focused(): # paranoia
      if self.last_selected_index != self.scroll_widget.get_selected_sprite_index():
        manager.playlist.play(manager.view.scroll_widget.get_focused_sprite_index())
      else: manager.show_nowplaying()

  def _callback(self):
    self._play_continuous()

  def update_selection(self):
    print "now update: %s" % (manager.playlist.current,)
    if self.dirty: self._load_scroll_widget()
    manager.nowplaying_button.update()

    if not manager.playlist.list or manager.playlist.is_empty():
      return

    list = self.scroll_widget.get_list()
    c = 0
    for row in list:
      if not manager.playlist.is_empty() and c == manager.playlist.current:
        self.scroll_widget.set_focused_index(c,snap=True)
        self.scroll_widget.set_selected_index(c)
        print "selectednow: %s" % (self.scroll_widget.get_selected_sprite().path,)
      c += 1

    self.last_selected_index = self.scroll_widget.get_selected_sprite_index()

  def remove_item(self):
    i = self.scroll_widget.get_focused_sprite_index()
    manager.playlist.remove_item(i)
    if manager.playlist.is_empty():
      manager.back()
      return False
    self.scroll_widget.calc_scrollable()
    self.scroll_widget.flush_visible()
    self.update_selection()

  def handle_event(self,event):
    if manager.nowplaying_button.is_pressed():
      if event.type == KEYDOWN and (event.key==K_UP or event.key==K_DOWN or event.key==K_RETURN or event.key==K_ESCAPE or event.key==K_F4):
        can_handle = (time.time()-self.combo_timeout_start > 0.1)

        if can_handle==True and event.key == K_UP and self.scroll_widget.has_focused():
          i = self.scroll_widget.get_focused_sprite_index()
          if i>0:
            manager.playlist.swap_items(i, i-1)
            self.scroll_widget.swapped_items(i, i-1)
        elif can_handle==True and event.key == K_DOWN and self.scroll_widget.has_focused():
          i = self.scroll_widget.get_focused_sprite_index()
          l = len(self.scroll_widget.get_list())-1
          if i<l:
            manager.playlist.swap_items(i, i+1)
            self.scroll_widget.swapped_items(i, i+1)
#        elif can_handle==True and event.key == K_RETURN and self.scroll_widget.has_focused():
#          self.remove_item()
        elif can_handle==True and event.key == K_RETURN:
          manager.playlist.save_as_dlg()
        elif can_handle==True and event.key == K_ESCAPE and self.scroll_widget.has_focused():
          i = self.scroll_widget.get_focused_sprite_index()
          if i>0:
            manager.playlist.swap_items(0, i)
            self.scroll_widget.swapped_items(0, i)
          l = len(self.scroll_widget.get_list())-1
          for j in xrange(1, l+1):
            manager.playlist.remove_item(1)
          self.scroll_widget.calc_scrollable()
          self.scroll_widget.flush_visible()
          self.update_selection()
        elif can_handle==True and event.key == K_F4 and self.scroll_widget.has_focused():
          i = self.scroll_widget.get_focused_sprite_index()
          song = self.scroll_widget.visible_sprites[i]['sprite']
          print "artist = ",song.artist
          manager.ignore_keypad_till = time.time() + 2
          manager.nowplaying_button.set_pressed(False) # will change views below
          manager.show_artist_albums_ext(song.artist_id,song.artist)

        manager.nowplaying_button.set_ignore_next_mouseup(True)
        self.combo_timeout_start = time.time()
        return True

    return BaseView.handle_event(self,event)


class VolumeView(TimerView):
  def __init__(self):
    ''' This is the nowplaying selection view '''
    TimerView.__init__(self,caption='Volume')
    self.album         = None
    self.cur_percent   = manager.player.volume
    self.slider_widget = None
    self._load_slider_widget()
    self.widgets.append(widgets.VolumeButton())
    self.update_selection()

  def _load_slider_widget(self):
    x = 130

    # Create Volume + Button
    plus_button = widgets.Button(theme.get_image('add'),self._plus_callback,True,repeat=True)
    plus_button.rect.left = x
    plus_button.rect.top  = manager.screen.get_rect().top + 18
    self.widgets.append(plus_button)

    # Create Volume - Button
    minus_button = widgets.Button(theme.get_image('remove'),self._minus_callback,True,repeat=True)
    minus_button.rect.left = x
    minus_button.rect.top  = manager.screen.get_rect().top + 420
    self.widgets.append(minus_button)

    # Create Slider Widget
    widget = widgets.Slider(Rect(
      x,
      manager.screen.get_rect().top + 60,
      57,
      360),
      self._slider_callback)
    
    self.slider_widget = widget
    self.widgets.append(widget)

  def _slider_callback(self):
    self.cur_percent = self.slider_widget.get_percent()
    print "volume callback: %s" % (self.cur_percent,)
    manager.player.set_volume(self.cur_percent)

  def _plus_callback(self):
    manager.player.volume_increase()

  def _minus_callback(self):
    manager.player.volume_decrease()

  def handle_event(self,event):
    if self.slider_widget:
      if event.type == KEYDOWN:
        if event.key == K_UP:
          self._plus_callback()
        elif event.key == K_DOWN:
          self._minus_callback()

    for widget in self.widgets:
      if widget.handle_event(event): return True
    return False

  def update(self):
    TimerView.update(self)
    if not self.slider_widget.is_idle(): self.reset_timeout()

  def update_selection(self):
    print "volume update: %s" % (manager.player.volume,)
    self.cur_percent = manager.player.volume
    self.slider_widget.set_percent(self.cur_percent)


class SongPositionView(BaseView):
  def __init__(self):
    ''' This is the nowplaying selection view '''
    BaseView.__init__(self,caption='Seek')
    self.album         = None
    self.cur_seconds   = manager.player.seconds
    self.step          = 10
    self._load_scroll_widget()
    self.widgets.append(widgets.VolumeButton())
    self.update()

  def _load_scroll_widget(self):
    if self.scroll_widget != None:
      self.widgets.remove(self.scroll_widget)
      self.scroll_widget = None
    gc.collect()

    manager.clear()
    status_area = widgets.Text("LOADING SONG POSITION")
    status_area.rect.topleft = (130,200)
    status_area.update()
    status_area.draw(manager.screen)
    pygame.display.flip()

    widget = widgets.ScrollWidget(Rect(
      130,
      manager.screen.get_rect().top,
      545,
      manager.screen.get_rect().bottom
    ),self._make_ok_button_callback(),
    self._get_unit_list(),
    self.time_generator_callback)

    self.scroll_widget = widget
    self.widgets.append(widget)

  def _calc_step(self):
    if manager.player.length > 3600:
      # 3600 = 1 hr in seconds
      self.step = 60
    else: self.step = 10

  def _get_unit_list(self):
    self._calc_step()
    l = range(0,int(manager.player.length),int(self.step))
    return l

  def time_generator_callback(self,id,seconds):
    return widgets.Time(seconds)

  def _make_ok_button_callback(self):
    def _ok_button_callback():
      sel_sprite = manager.view.scroll_widget.get_focused_sprite()
      if self.cur_seconds == sel_sprite.seconds:
        manager.back()
      else:
        manager.player.seek(sel_sprite.seconds,2)
        self.update_selection(False,sel_sprite.seconds)
    return _ok_button_callback

  def update_selection(self,reload=True,seconds=None):
    print "songposition_view update:"
    if reload: self._load_scroll_widget()
    if not seconds: seconds = manager.player.seconds
    if seconds > 0:
      self.cur_seconds = int(round(seconds/self.step) * self.step)
    else:
      self.cur_seconds = 0
    print "cur_seconds = %s" % (self.cur_seconds,)
    list = self.scroll_widget.get_list()
    c = 0
    for row in list:
      if row == self.cur_seconds:
        if not (self.scroll_widget.get_focused_sprite() and self.scroll_widget.get_focused_sprite().seconds == self.cur_seconds):
          self.scroll_widget.set_focused_index(c,snap=True)
          self.scroll_widget.set_selected_index(c)
        print "selected songpos: %s" % (self.scroll_widget.get_selected_sprite().seconds,)
      c += 1


class ExitConfirmView(BaseView):
  def __init__(self):
    ''' This is the exit confirmation dialog/view '''
    BaseView.__init__(self)
    self.album          = None
    self.add_to_history = False
    self._load_ok_button()
    self._load_cancel_button()
    self.widgets.append(widgets.VolumeButton())
    status_area = widgets.Text("                     Quit Kagu?         ")
    status_area.rect.topleft = (130,200)
    status_area.update()
    self.widgets.append(status_area)

  def _make_ok_button_callback(self):
    def _ok_button_callback():
      manager.quit()
      sys.exit(0)
    return _ok_button_callback

  def _make_cancel_button_callback(self):
    def _cancel_button_callback():
      manager.back()
    return _cancel_button_callback

  def _load_ok_button(self):
    rect = Rect(200,300,75,75)
    image = theme.get_surface((rect.width,rect.height))
    image.blit(manager.background,(0,0),rect)
    globals.gprint("OK",image,(25,25))
    button = widgets.Button(image,self._make_ok_button_callback(),True)
    button.rect.left = rect.left
    button.rect.top  = rect.top
    self.widgets.append(button)

  def _load_cancel_button(self):
    rect = Rect(400,300,200,75)
    image = theme.get_surface((rect.width,rect.height))
    image.blit(manager.background,(0,0),rect)
    globals.gprint("CANCEL",image,(25,25))
    button = widgets.Button(image,self._make_cancel_button_callback(),False)
    button.rect.left = rect.left
    button.rect.top  = rect.top
    self.widgets.append(button)


class AlbumCoverView(BaseView):
  def __init__(self):
    ''' This is the Zoomed Album Cover view '''
    BaseView.__init__(self, 'Album Cover')
    self.album         = None
    self.album_widget  = None
    self.sync_with_nowplaying = False
    self.album_id      = None
    self.nowplaying    = None
    self.widgets.append(widgets.VolumeButton())
    self.update()

  def _load_album_widget(self,album_id=None):
    if not album_id:
      if not manager.playlist.list or manager.playlist.is_empty(): return
      if self.nowplaying == manager.playlist.status(): return # don't run DB query if nothing has changed
      self.nowplaying = manager.playlist.status()
      processtart=time.time()
      album_id = db.song_of_path(manager.playlist.status())[0]['album_id']
      print "album cover DB query (song_of_path) took:" + str(time.time()-processtart)
    else: self.nowplaying = None

    if self.album_id != None and self.album_id == album_id: return # don't update if nothing has changed

    if self.album_widget: # clean up old widget
      self.widgets.remove(self.album_widget)
      self.album_widget = None

    self.album_id = album_id
    processtart=time.time()
    art_path = db.art_path_of_album(album_id)
    art = widgets.CoverArt(album_id,530,manager.screen.get_rect().bottom,art_path)
    self.album_widget = widgets.Button(art.image,self._callback,is_default=True)
    # center widget on screen
    self.album_widget.rect.top  = manager.screen.get_rect().centery - self.album_widget.rect.height / 2
    self.album_widget.rect.left = manager.screen.get_rect().centerx - self.album_widget.rect.width  / 2
    print "load album cover took:" + str(time.time()-processtart)
    self.widgets.append(self.album_widget)

  def _callback(self):
    if self.sync_with_nowplaying: return manager.show_nowplaying()
    else: return manager.back()

  def update_image(self,album_id=None,sync_with_nowplaying=False):
    print "AlbumCover update:"
    self.sync_with_nowplaying = sync_with_nowplaying
    self._load_album_widget(album_id)

  def update(self):
    if self.sync_with_nowplaying: self._load_album_widget()
    BaseView.update(self)


class ArtistCoverView(BaseView):
  def __init__(self):
    ''' This is the Zoomed Artist Cover view '''
    BaseView.__init__(self, 'Artist Image')
    self.artist        = None
    self.artist_widget = None
    self.artist_id     = None
    self.widgets.append(widgets.VolumeButton())
    self.update()

  def _load_artist_widget(self,artist_id=None):
    if self.artist_id != None and self.artist_id == artist_id: return # don't update if nothing has changed

    if self.artist_widget: # clean up old widget
      self.widgets.remove(self.artist_widget)
      self.artist_widget = None

    self.artist_id = artist_id
    processtart=time.time()
    art_path = db.art_path_of_artist(artist_id)
    art = widgets.CoverArt(artist_id,530,manager.screen.get_rect().bottom,art_path)
    self.artist_widget = widgets.Button(art.image,manager.back)
    # center widget on screen
    self.artist_widget.rect.top  = manager.screen.get_rect().centery - self.artist_widget.rect.height / 2
    self.artist_widget.rect.left = manager.screen.get_rect().centerx - self.artist_widget.rect.width  / 2
    print "load artist cover took:" + str(time.time()-processtart)
    self.widgets.append(self.artist_widget)

  def update_image(self,artist_id=None):
    print "ArtistCover update:"
    self._load_artist_widget(artist_id)


class MenuView(BaseView):
  def __init__(self):
    ''' This is the main menu view '''
    BaseView.__init__(self, 'Kagu')
    self.album_widget = None
    self._load_album_widget()
    self.scroll_widget = self.album_widget
    self.widgets.append(widgets.VolumeButton())

  def _load_album_widget(self):
    if self.album_widget: return

    manager.clear()
    status_area = widgets.Text("LOADING MENU")
    status_area.rect.topleft = (130,200)
    status_area.update()
    status_area.draw(manager.screen)
    pygame.display.flip()

    widget = widgets.ScrollWidget(Rect(
      130,
      manager.screen.get_rect().top,
      545,
      manager.screen.get_rect().bottom
    ),self._callback,
    self._get_menu_list(),
    self.menu_generator_callback,
    wrap=False)
    
    self.album_widget = widget
    self.widgets.append(widget)

  def _get_menu_list(self):
    list = ['Albums','Artists','All Songs','Playlists','Now Playing','Sleep Timer']
    if globals.ISMAEMO:
      list.append('Turn Off Screen')
      if manager.has_mplayer:
        list.append('Player')
    list.append('Scrollbars')
    list.append('Toggle Fullscreen')
    return list

  def menu_generator_callback(self,id,row):
    if row=='Sleep Timer':
      if manager.sleep_timer>0:
        if manager.sleep_timer_dur == 3600:
          row+=': 1 Hour'
        elif manager.sleep_timer_dur == 7200:
          row+=': 2 Hours'
        else:
          row+=': ON?'
      else:
        row+=': OFF'
    elif row=='Player':
      if manager.default_player=='ossoplayer':
        row+=': OSSOPlayer'
      elif manager.default_player=='mplayer':
        row+=': MPlayer'
      else:
        row+=': Unknown'
    elif row=='Scrollbars':
      if manager.scrollbar_state:
        row+=': ON'
      else:
        row+=': OFF'
      
    return widgets.Text(row)

  def _callback(self):
    update = False
    sel_sprite_index = manager.view.scroll_widget.get_focused_sprite_index()
    name = manager.view.scroll_widget.get_list()[sel_sprite_index]
    if   name == 'Albums':      manager.show_albums()
    elif name == 'All Songs':   manager.show_all_songs()
    elif name == 'Artists':     manager.show_artists()
    elif name == 'Playlists':   manager.show_playlists()
    elif name == 'Now Playing': manager.show_nowplaying()
    elif name == 'Toggle Fullscreen': pygame.display.toggle_fullscreen()
    elif name == 'Scrollbars':
      manager.scrollbar_state = not manager.scrollbar_state
      if manager.view.scroll_widget:
        manager.view.scroll_widget.scrollbar_toggle(manager.scrollbar_state)
      update = True

    elif name == 'Sleep Timer':
      if manager.sleep_timer>0:
        if manager.sleep_timer_dur==3600:
          manager.sleep_timer_dur=7200
        else:
          manager.sleep_timer=0
      else:
        manager.sleep_timer=1
        manager.sleep_timer_dur=3600
      update = True
    elif name == 'Player':
      if manager.default_player=='mplayer':
        manager.default_player='ossoplayer'
      else:
        manager.default_player='mplayer'
      manager.switch_player()
      update = True
        
    elif name == 'Turn Off Screen':
      globals.blank_screen()
      manager.screen_manually_blanked = True

    if update: self.scroll_widget.swapped_items(0, 0)


